"""The FastAnchor wraps the fast-anchor component"""
from bokeh.core import properties
from bokeh.models import Widget


class FastAnchor(Widget):
    """Bokeh model wrapping the fast-anchor component"""

    appearance = properties.String()
    download = properties.String()
    href = properties.String()
    hreflang = properties.String()
    ping = properties.String()
    referrerpolicy = properties.String()
    referrer = properties.String()  # cannot call this ref
    rel = properties.String()
    target = properties.String()
    mimetype = properties.String()



"""The FastGalleryTemplate can be used to show case your applications and other resources in a
nice Gallery"""
import pathlib

import param
from panel.template.base import FAVICON_URL, Template

from awesome_panel_extensions.site.models import Application

ROOT = pathlib.Path(__file__).parent
CSS = (ROOT / "gallery_template.css").read_text()
JS = (ROOT / "gallery_template.js").read_text()
TEMPLATE = (ROOT / "gallery_template.html").read_text()


class GalleryTemplate(Template):
    """The GalleryTemplate can be used to show case your applications in a nice way"""

    applications = param.List(
        class_=Application,
        doc="""
        The list of applications to show in the Gallery
        """,
    )
    site = param.String(
        default="Panel",
        doc="""
        The name of the site. Will be shown in the header and link to the
        site_url. Default is 'Panel'.""",
    )
    site_url = param.String(
        "/",
        doc="""
        Url of the site and logo. Default is '/'.""",
    )
    title = param.String(
        default="Gallery",
        doc="""
        The name of the gallery. Shown in the header and browser tab. Default is 'Gallery'""",
    )
    description = param.String(
        "Applications running on the server",
        doc="""
    A description shown in the header. Default is 'Applications running on the server'.
    """,
    )

    background_image = param.String(
        "https://preview.redd.it/9oi428ohy7t21.png?auto=webp&s=5051b77d33e85446b6492a1e02725c6729777d4f",  # pylint: disable=line-too-long
        doc="""
        The url of an image to show in the header""",
    )

    theme = param.ObjectSelector("default", objects=["dark", "default"])
    theme_toggle = param.Boolean(
        default=True,
        doc="""
        Whether or not to enable toggling the theme via a switchbox. Default is True""",
    )
    font_family = param.String(
        "Open Sans",
        doc="""
    The name of the font family to use. Default is 'Open Sans'.""",
    )
    font_url = param.String("//fonts.googleapis.com/css?family=Open+Sans")
    accent_base_color = param.String(
        "#E1477E",
        doc="""
    The Accent Color. Default is '#E1477E'.""",
    )
    footer = param.String(
        doc="""
    Footer text. Default is ''"""
    )

    target = param.ObjectSelector(
        "_self",
        objects=["_blank", "_parent", "_top", "_self"],
        doc="""
    How to open the url. One of '_blank', '_parent', '_top' or '_self' (default).""",
    )

    favicon = param.String(
        FAVICON_URL,
        constant=True,
        doc="""
        URI of favicon to add to the document head (if local file, favicon is
        base64 encoded as URI).""",
    )
    meta_name = param.String(
        doc="""
        A meta name to add to the document head for search
        engine optimization. For example 'HoloViz Panel Gallery'. If none is specified the title
        will be used."""
    )
    meta_description = param.String(
        doc="""
        A meta description to add to the document head for search
        engine optimization. For example 'Applications running on the server'. If none is specified
        the description will be used."""
    )
    meta_keywords = param.String(
        doc="""
        Meta keywords to add to the document head for search engine
        optimization. For example 'HoloViz, Panel, Gallery'. Default is ''."""
    )
    meta_author = param.String(
        doc="""
        A meta author to add to the the document head for search
        engine optimization. For example 'HoloViz Panel'."""
    )
    meta_viewport = param.String(
        doc="""
        A meta viewport to add to the header."""
    )

    def __init__(self, **params):  # pylint: disable=too-many-arguments
        """The FastGalleryTemplate can be used to show case your applications in a nice way"""
        super().__init__(template=TEMPLATE, **params)
        self.add_variable("site_name", self.site)
        self.add_variable("site_url", self.site_url)
        self.add_variable("name", self.title)
        self.add_variable("url", "")
        self.add_variable("description", self.description)
        self.add_variable("background_image", self.background_image)
        self.add_variable("applications", list(sorted(self.applications, key=lambda x: x.name)))
        self.add_variable("gallery_js", JS)
        self.add_variable("gallery_css", CSS)
        self.add_variable("app_favicon", self.favicon)
        self.add_variable("target", self.target)
        self.add_variable("theme", self.theme)
        self.add_variable("theme_toggle", self.theme_toggle)
        self.add_variable("font_family", self.font_family)
        self.add_variable("font_url", self.font_url)
        self.add_variable("accent_base_color", self.accent_base_color)

        self.add_variable("head_title", self._head_title)
        self.add_variable("meta_name", self.meta_name or self._head_title)
        self.add_variable("meta_description", self.meta_description or self.description)
        self.add_variable("meta_keywords", self.meta_keywords)
        self.add_variable("meta_author", self.meta_author)
        self.add_variable("meta_viewport", self.meta_viewport)

        self.add_variable("footer", self.footer)

    @property
    def _head_title(self):
        if self.site and self.title:
            return self.site + " | " + self.title
        if self.title:
            return self.title
        if self.site:
            return self.site
        return "Panel Gallery"



"""The FastGalleryTemplate can be used to show case your applications in a nice way"""
import pathlib

import param
from panel import Template

ROOT = pathlib.Path(__file__).parent
CSS = (ROOT / "fast_gallery_template.css").read_text()
JS = (ROOT / "fast_gallery_template.js").read_text()
TEMPLATE = (ROOT / "fast_gallery_template.html").read_text()
FAVICON = (
    "https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel/"
    "2781d86d4ed141889d633748879a120d7d8e777a/assets/images/favicon.ico"
)


class FastGalleryTemplate(Template):
    """The FastGalleryTemplate can be used to show case your applications in a nice way"""

    resources = param.List()
    title = param.String(
        default="Gallery",
        doc="""
        A title to show in the header. Also added to the document head
        meta settings and as the browser tab title.""",
    )
    description = param.String("Applications running on the server")
    site = param.String(
        default="",
        doc="""
        The name of the site. Will be shown in the header and link to the
        root of the site. Default is '', i.e. not shown.""",
    )
    site_url = param.String(
        "/",
        doc="""
        Url of the site and logo. Default is '/'.""",
    )

    meta_name = param.String(
        doc="""
        A meta name to add to the document head for search
        engine optimization. For example 'HoloViz Panel Gallery'."""
    )
    meta_description = param.String(
        doc="""
        A meta description to add to the document head for search
        engine optimization. For example 'Applications running on the server'."""
    )
    meta_keywords = param.String(
        doc="""
        Meta keywords to add to the document head for search engine
        optimization. For example 'HoloViz, Panel, Gallery'"""
    )
    meta_author = param.String(
        doc="""
        A meta author to add to the the document head for search
        engine optimization. For example 'HoloViz Panel'."""
    )
    meta_viewport = param.String(
        doc="""
        A meta viewport to add to the header."""
    )

    background_image_url = param.String(
        "https://preview.redd.it/9oi428ohy7t21.png?auto=webp&s=5051b77d33e85446b6492a1e02725c6729777d4f"  # pylint: disable=line-too-long
    )
    target = param.ObjectSelector("_self", objects=["_blank", "_parent", "_top", "_self"])
    favicon = param.String(
        FAVICON,
        constant=True,
        doc="""
        URI of favicon to add to the document head (if local file, favicon is
        base64 encoded as URI).""",
    )
    theme = param.ObjectSelector("default", objects=["dark", "default"])
    theme_toggle = param.Boolean(
        default=True,
        doc="""
        If True a switch to toggle the Theme is shown.""",
    )
    font_family = param.String("Open Sans")
    font_url = param.String("//fonts.googleapis.com/css?family=Open+Sans")
    accent_base_color = param.String("#E1477E")

    def __init__(self, **params):  # pylint: disable=too-many-arguments
        """The FastGalleryTemplate can be used to show case your applications in a nice way"""
        super().__init__(template=TEMPLATE, **params)
        self.add_variable("site_name", self.site)
        self.add_variable("site_url", self.site_url)
        self.add_variable("name", self.title)
        self.add_variable("url", "")
        self.add_variable("description", self.description)
        self.add_variable("background_image_url", self.background_image_url)
        self.add_variable("items", list(sorted(self.resources, key=lambda x: x.name)))
        self.add_variable("gallery_js", JS)
        self.add_variable("gallery_css", CSS)
        self.add_variable("app_favicon", self.favicon)
        self.add_variable("target", self.target)
        self.add_variable("theme", self.theme)
        self.add_variable("theme_toggle", self.theme_toggle)
        self.add_variable("font_family", self.font_family)
        self.add_variable("font_url", self.font_url)
        self.add_variable("accent_base_color", self.accent_base_color)

        self.add_variable("head_title", self._head_title)
        self.add_variable("meta_name", self.meta_name or self._head_title)
        self.add_variable("meta_description", self.meta_description or self.description)
        self.add_variable("meta_keywords", self.meta_keywords)
        self.add_variable("meta_author", self.meta_author)
        self.add_variable("meta_viewport", self.meta_viewport)

    @property
    def _head_title(self):
        if self.site and self.title:
            return self.site + " | " + self.title
        if self.title:
            return self.title
        if self.site:
            return self.site
        return "Panel Gallery"

    # def create_from_toml(path: Union[str, pathlib.Path]) -> "FastGalleryTemplate":
    #     raise NotImplementedError()



"""The FastAnchor enables the user to click a link and navigate to it.

The component supports several visual apperances
(accent, lightweight, neutral, outline, stealth, hypertext).

The FastAnchor wraps the `fast-anchor` of the [Fast Design](https://fast.design/) Framework.

For more information view the [component specification]\
(https://github.com/microsoft/fast/tree/master/packages/web-components/fast-foundation/\
src/anchor/anchor.spec.md).

See also https://explore.fast.design/components/fast-anchor.
    """
import param  # pylint: disable=wrong-import-order
from panel.widgets import Widget

from awesome_panel_extensions.bokeh_extensions.fast.fast_anchor import FastAnchor as _BkFastAnchor

FAST_ANCHOR_APPEARENCES = [
    "accent",
    "lightweight",
    "neutral",
    "outline",
    "stealth",
    "hypertext",
]
# pylint: disable=line-too-long
RELS = [
    None,
    "alternate",  # Provides a link to an alternate representation of the document (i.e. print page, translated or mirror)
    "author",  # Provides a link to the author of the document
    "bookmark",  # Permanent URL used for bookmarking
    "external",  # Indicates that the referenced document is not part of the same site as the current document
    "help",  # Provides a link to a help document
    "license",  # Provides a link to licensing information for the document
    "next",  # Provides a link to the next document in the series
    "nofollow",  # Links to an unendorsed document, like a paid link.
    "noreferrer",  # Requires that the browser should not send an HTTP referer header if the user follows the hyperlink
    "noopener",  # Requires that any browsing context created by following the hyperlink must not have an opener browsing context
    "prev",  # The previous document in a selection
    "search",  # Links to a search tool for the document
    "tag",  # A tag (keyword) for the current document
]
# pylint: enable=line-too-long
TARGETS = [
    None,
    "_blank",
    "_parent",
    "_self",
    "_top",
]
REFERRER_POLICIES = [
    None,
    "no-referrer",
    "no-referrer-when-downgrade",
    "origin",
    "origin-when-cross-origin",
    "same-origin",
    "strict-origin",
    "strict-origin-when-cross-origin",
    "unsafe-url",
]


class FastAnchor(Widget):
    """The FastAnchor enables the user to click a link and navigate to it.

The component supports several visual apperances
(accent, lightweight, neutral, outline, stealth, hypertext).

When using the `FastTemplate` you can also use the `<fast-anchor>` tag directly inside
`pn.pane.Markdown` and `pn.pane.HTML`.

The FastAnchor wraps the `fast-anchor` of the [Fast Design](https://fast.design/) Framework.

For more information view the [component specification]\
(https://github.com/microsoft/fast/tree/master/packages/web-components/fast-foundation/\
src/anchor/anchor.spec.md).

See also https://explore.fast.design/components/fast-anchor.
    """

    value = param.String(
        default=None,
        allow_None=True,
        doc="""The URL that the hyperlink points to. Default is None.""",
        label="Href",
    )
    appearance = param.ObjectSelector(
        default=None,
        objects=FAST_ANCHOR_APPEARENCES,
        doc="""Determines the appearance of the anchor. One of `accent`, `lightweight`, `neutral`,
        `outline`, `stealth` or `hypertext`. Defaults to None/ neutral""",
        allow_None=True,
    )
    target = param.ObjectSelector(
        default=None,
        objects=TARGETS,
        allow_None=True,
        doc="""Where to display the linked URL. One of None, `_self`, `_blank`, `_parent`, `_self`
        or `_top`. Defaults to None""",
    )

    download = param.String(
        default=None,
        allow_None=True,
        doc="""Prompts the user to save the linked URL instead of navigating to it.
        Can be used with or without a value. Defaults to None""",
    )
    # We call this value instead of href in order to be able to use it
    # with pn.Param
    hreflang = param.String(
        default=None,
        allow_None=True,
        doc="""Hints at the human language of the linked URL. No built-in functionality.
        Default is None.""",
    )
    ping = param.String(
        default=None,
        allow_None=True,
        doc="""A space-separated list of URLs. When the link is followed, the browser will send
        POST requests with the body PING to the URLs. Typically for tracking.
        Default is None.""",
    )
    referrerpolicy = param.ObjectSelector(
        default=None,
        objects=REFERRER_POLICIES,
        allow_None=True,
        doc="""How much of the referrer to send when following the link.
        one of no-referrer, no-referrer-when-downgrade, origin,origin-when-cross-origin,
        same-origin, strict-origin, strict-origin-when-cross-origin and unsafe-url.check_on_set.
        Defaults to None""",
    )
    rel = param.ObjectSelector(
        default=None,
        objects=RELS,
        allow_None=True,
        doc="""The relationship of the linked URL as space-separated link types like alternate,
        archives, ... See https://developer.mozilla.org/en-US/docs/Web/HTML/Link_types.
        Defaults to None""",
    )
    mimetype = param.String(
        default=None,
        allow_None=True,
        doc="""Hints at the linked URL’s format with a MIME type. No built-in functionality.
        Default is None.""",
        label="Type",
    )

    height = param.Integer(default=40, bounds=(0, None))

    _widget_type = _BkFastAnchor

    _rename = {
        "value": "href",
        "ref": "referrer",  # pylint: disable=protected-access
    }

    def __init__(self, **params):
        super().__init__(**params)

        self._set_height()

    @param.depends("appearance", watch=True)
    def _set_height(self, *_):
        if self.appearance == "hypertext":
            self.height = 20
        else:
            self.height = 40



"""Base Progress WebComponent"""
import param

from awesome_panel_extensions.web_component import WebComponent


class Progress(WebComponent):
    """Base Progress WebComponent"""

    active = param.Boolean(
        default=True,
        doc="""
        If no value is set the active property toggles animation of the
        progress bar on and off.""",
    )

    bar_color = param.ObjectSelector(
        default="success",
        objects=["primary", "secondary", "success", "info", "danger", "warning", "light", "dark"],
    )

    max = param.Integer(default=100, doc="The maximum value of the progress bar.")

    value = param.Integer(
        default=None,
        bounds=(0, 100),
        doc="""
        The current value of the progress bar. If set to None the progress
        bar will be indeterminate and animate depending on the active
        parameter.""",
    )

    @param.depends("max", watch=True)
    def _update_value_bounds(self):
        self.param.value.bounds = (0, self.max)



# pylint: disable=redefined-outer-name,protected-access
# pylint: disable=missing-function-docstring,missing-module-docstring,missing-class-docstring
# pylint: disable=too-many-statements, not-callable, too-complex, too-many-instance-attributes
# pylint: disable=too-many-branches
import inspect
import math
from datetime import date, datetime

import holoviews as hv
import numpy as np
import pandas as pd
import panel as pn
import param
from holoviews import opts
from panel import widgets as pnw
from panel.template import FastListTemplate

from awesome_panel_extensions.widgets.dataframe import get_default_formatters

pn.extension("echarts", "ace")
hv.extension("bokeh")

ACCENT = "#1f77b4"
ACCENT_REST = ACCENT


def get_dataframe():
    return pd.DataFrame(np.random.randint(0, 100, size=(25, 4)), columns=list("ABCD"))


PANES = [
    pn.pane.HoloViews,
    pn.pane.ECharts,
]
WIDGETS = [
    pnw.Ace,
    pnw.AutocompleteInput,
    pnw.Button,
    pnw.Checkbox,
    pnw.CheckBoxGroup,
    pnw.CheckButtonGroup,
    pnw.ColorPicker,
    pnw.CrossSelector,
    pnw.DataFrame,
    pnw.DatePicker,
    pnw.DateRangeSlider,
    pnw.DateSlider,
    pnw.DatetimeInput,
    pnw.DatetimeRangeInput,
    pnw.DiscretePlayer,
    pnw.DiscreteSlider,
    pnw.FileDownload,
    pnw.FileInput,
    pnw.FileSelector,
    pnw.FloatInput,
    pnw.FloatSlider,
    pnw.IntInput,
    pnw.IntRangeSlider,
    pnw.IntSlider,
    pnw.LiteralInput,
    pnw.MenuButton,
    pnw.MultiChoice,
    pnw.MultiSelect,
    pnw.PasswordInput,
    pnw.Player,
    pnw.Progress,
    pnw.RadioBoxGroup,
    pnw.RadioButtonGroup,
    pnw.RangeSlider,
    pnw.Select,
    pnw.StaticText,
    pnw.TextAreaInput,
    pnw.TextInput,
    pnw.Toggle,
    # pnw.VideoStream,
]
LAYOUTS = [pn.layout.Divider]

COMPONENTS = {"Layout": LAYOUTS, "Pane": PANES, "Widget": WIDGETS}
DEFAULT_COMPONENT_TYPE = "Pane"
DEFAULT_COMPONENT = {
    "Layout": pn.layout.Divider,
    "Pane": pn.pane.HoloViews,
    "Widget": pn.widgets.DataFrame,
}


def _create_hvplot():
    # Generate some data
    cl1 = np.random.normal(loc=2, scale=0.2, size=(200, 200))
    cl2x = np.random.normal(loc=-2, scale=0.6, size=200)
    cl2y = np.random.normal(loc=-2, scale=0.1, size=200)
    cl3 = np.random.normal(loc=0, scale=1.5, size=(400, 400))
    # Create an overlay of points and ellipses
    clusters = (
        hv.Points(cl1).opts(color="blue")
        * hv.Points((cl2x, cl2y)).opts(color="green")
        * hv.Points(cl3).opts(color="#FDDC22")
    ).opts(opts.Points(tools=["hover"]))
    plot = (
        clusters
        * hv.Ellipse(2, 2, 2).opts(line_width=3, color=ACCENT_REST)
        * hv.Ellipse(-2, -2, (4, 2)).opts(line_width=3, color=ACCENT_REST)
    )
    return plot


def _create_echarts_plot():
    echart = {
        "tooltip": {},
        "legend": {"data": ["Sales"]},
        "xAxis": {
            "data": ["shirt", "cardign", "chiffon shirt", "pants", "heels", "socks"],
            "axisLine": {"lineStyle": {"color": "#ccc"}},
        },
        "yAxis": {
            "axisLine": {"lineStyle": {"color": "#ccc"}},
        },
        "series": [
            {
                "name": "Sales",
                "type": "bar",
                "data": [
                    1.0,
                    1.2,
                    1.4,
                    1.6,
                    1.8,
                    2.0,
                ],
                "itemStyle": {"color": "#DF3874"},
            }
        ],
        "responsive": True,
    }
    text_style = {"color": "#ccc"}
    update = ["legend", "xAxis", "yAxis"]
    for upd in update:
        echart[upd]["textStyle"] = text_style
    return echart


class PanelComponentExplorer(param.Parameterized):
    component_type = param.ObjectSelector(
        DEFAULT_COMPONENT_TYPE, objects=list(COMPONENTS.keys()), label="Type"
    )
    component = param.ObjectSelector(
        DEFAULT_COMPONENT[DEFAULT_COMPONENT_TYPE], COMPONENTS[DEFAULT_COMPONENT_TYPE]
    )
    update = param.Action()

    def __init__(self, **params):
        super().__init__(**params)
        self._default_component = DEFAULT_COMPONENT.copy()
        self.view = self._create_view()
        self._update_widgets_panel()

    def _create_view(self):
        self._css_panel = pn.pane.HTML(height=0, width=0, margin=0, sizing_mode="fixed")
        self._settings_panel = pn.Column(
            pn.pane.Markdown("## Selections"),
            pn.Param(
                self,
                parameters=["component_type", "component"],
                expand_button=False,
                show_name=False,
            ),
            self._css_panel,
        )
        self._component_panel = pn.Column()

        self._template = FastListTemplate(
            site="Awesome Panel",
            title="Component Explorer",
            sidebar=[self._settings_panel],
            main=[self._component_panel],
            main_max_width="1024px",
            header_background=ACCENT,
            accent_base_color=ACCENT,
        )
        if "Dark" in str(self._template.theme):
            self._theme = "dark"
            self._ace_theme = "tomorrow_night"
        else:
            self._theme = "default"
            self._ace_theme = "tomorrow"
        return self._template

    @pn.depends("component_type", watch=True)
    def _update_component_list(self):
        self.param.component.objects = COMPONENTS[self.component_type]
        self.component = self._default_component[self.component_type]

    @pn.depends("component", watch=True)
    def _update_widgets_panel(self):
        self._default_component[self.component_type] = self.component

        component = None
        controls = None
        if self.component is pn.pane.HoloViews:
            component = pn.pane.HoloViews(_create_hvplot())
        if self.component is pn.pane.ECharts:
            # Issue https://github.com/holoviz/panel/issues/1817
            component = pn.pane.ECharts(
                _create_echarts_plot(), min_height=400, min_width=200, sizing_mode="stretch_both"
            )
        if self.component is pnw.Ace:
            py_code = inspect.getsource(_create_hvplot)
            component = pnw.Ace(
                value=py_code,
                sizing_mode="stretch_width",
                language="python",
                height=400,
                theme=self._ace_theme,
            )
        elif self.component is pnw.AutocompleteInput:
            component = pnw.AutocompleteInput(
                name="Autocomplete Input",
                options=["Biology", "Chemistry", "Physics"],
                placeholder="Write something here",
            )
        elif self.component is pnw.Button:
            component = pnw.Button(name="Click me", button_type="primary")
        elif self.component is pnw.CheckBoxGroup:
            component = pnw.CheckBoxGroup(
                name="Checkbox Group",
                value=["Apple", "Pear"],
                options=["Apple", "Banana", "Pear", "Strawberry"],
                inline=True,
            )
        elif self.component is pnw.CheckButtonGroup:
            component = pnw.CheckButtonGroup(
                name="Check Button Group",
                value=["Apple", "Pear"],
                options=["Apple", "Banana", "Pear", "Strawberry"],
                button_type="success",
            )
        elif self.component is pnw.Checkbox:
            component = pnw.Checkbox(name="Checkbox")
        elif self.component is pnw.ColorPicker:
            component = pnw.ColorPicker(name="Color Picker", value="#DF3874")
        elif self.component is pnw.CrossSelector:
            component = pnw.CrossSelector(
                name="Fruits",
                value=["Apple", "Pear"],
                options=["Apple", "Banana", "Pear", "Strawberry"],
                height=300,
            )
        elif self.component is pnw.DataFrame:
            component = self.component(name="Hello")
            component.value = get_dataframe()
            component.formatters = get_default_formatters(component.value)
            controls = pn.Spacer()
        elif self.component is pnw.DatePicker:
            component = pnw.DatePicker(name="Date Picker")
            # Issue: https://github.com/holoviz/panel/issues/1810
            # component.start = date(2020, 1, 20)
            # component.end = date(2020, 2, 20)
            # component.value = date(2020, 2, 18)
        elif self.component is pnw.DateRangeSlider:
            component = self.component(name="Hello")
            component.start = date(2020, 1, 20)
            component.end = date(2020, 2, 20)
            component.value = (date(2020, 2, 18), date(2020, 2, 20))
        elif self.component is pnw.DateSlider:
            component = self.component(name="Hello")
            component.start = date(2020, 1, 20)
            component.end = date(2020, 2, 20)
            component.value = date(2020, 2, 18)
        elif self.component is pnw.DatetimeInput:
            component = self.component(name="Hello")
            component.value = datetime(2020, 2, 18, 1, 2, 3)
        elif self.component is pnw.DatetimeRangeInput:
            component = self.component(
                name="Hello",
                start=datetime(2020, 1, 20),
                end=datetime(2020, 2, 20),
                value=(datetime(2020, 2, 18), datetime(2020, 2, 20)),
            )
        elif self.component is pnw.DiscretePlayer:
            component = pnw.DiscretePlayer(
                name="Discrete Player",
                options=[2, 4, 8, 16, 32, 64, 128],
                value=32,
                loop_policy="loop",
            )
        elif self.component is pnw.DiscreteSlider:
            component = pnw.DiscreteSlider(
                name="Discrete Slider", options=[2, 4, 8, 16, 32, 64, 128], value=32
            )
        elif self.component is pnw.FileDownload:
            component = pnw.FileDownload(file="README.md", filename="README.md")
        elif self.component is pnw.FileInput:
            component = pnw.FileInput(accept=".csv,.json")
        elif self.component is pnw.FileSelector:
            component = pnw.FileSelector(name="Hello", max_height=400)
        elif self.component is pnw.FloatInput:
            component = pnw.FloatInput(name="FloatInput", value=5.0, step=1e-1, start=0, end=1000)
        elif self.component is pnw.FloatSlider:
            component = pnw.FloatSlider(
                name="Float Slider", start=0, end=3.141, step=0.01, value=1.57
            )
        elif self.component is pnw.IntInput:
            component = pnw.IntInput(name="IntInput", value=5, step=2, start=0, end=1000)
        elif self.component is pnw.IntRangeSlider:
            component = pnw.IntRangeSlider(
                name="Integer Range Slider", start=0, end=100, value=(8, 40), step=2
            )
        elif self.component is pnw.IntSlider:
            component = pnw.IntSlider(name="Integer Slider", start=0, end=20, step=2, value=4)
        elif self.component is pnw.LiteralInput:
            component = pnw.LiteralInput(
                name="Literal Input (dict)", value={"key": [1, 2, 3]}, type=dict
            )
        elif self.component is pnw.MenuButton:
            menu_items = [
                ("Option A", "a"),
                ("Option B", "b"),
                ("Option C", "c"),
                None,
                ("Help", "help"),
            ]
            component = pnw.MenuButton(name="Dropdown", items=menu_items, button_type="primary")
        elif self.component is pnw.MultiChoice:
            component = pnw.MultiChoice(
                name="MultiSelect",
                value=["Apple", "Pear"],
                options=["Apple", "Banana", "Pear", "Strawberry"],
            )
        elif self.component is pnw.MultiSelect:
            component = pnw.MultiSelect(
                name="MultiSelect",
                value=["Apple", "Pear"],
                options=["Apple", "Banana", "Pear", "Strawberry"],
                size=8,
            )
        elif self.component is pnw.PasswordInput:
            component = pnw.PasswordInput(
                name="Password Input", placeholder="Enter a string here..."
            )
        elif self.component is pnw.Player:
            component = pnw.Player(name="Player", start=0, end=100, value=32, loop_policy="loop")
        elif self.component is pnw.Progress:
            component = pnw.Progress(name="Progress", value=20, width=200)
        elif self.component is pnw.RadioBoxGroup:
            component = pnw.RadioBoxGroup(
                name="RadioBoxGroup", options=["Biology", "Chemistry", "Physics"], inline=True
            )
        elif self.component is pnw.RadioButtonGroup:
            component = pnw.RadioButtonGroup(
                name="Radio Button Group",
                options=["Biology", "Chemistry", "Physics"],
                button_type="success",
            )
        elif self.component is pnw.RangeSlider:
            component = pnw.RangeSlider(
                name="Range Slider",
                start=0,
                end=math.pi,
                value=(math.pi / 4.0, math.pi / 2.0),
                step=0.01,
            )
        elif self.component is pnw.Select:
            component = pnw.Select(name="Select", options=["Biology", "Chemistry", "Physics"])
        elif self.component is pnw.StaticText:
            component = pnw.StaticText(name="Static Text", value="A string")
        elif self.component is pnw.TextAreaInput:
            component = pnw.input.TextAreaInput(
                name="Text Area Input", placeholder="Enter a string here..."
            )
        elif self.component is pnw.TextInput:
            component = pnw.TextInput(name="Text Input", placeholder="Enter a string here...")
        elif self.component == pnw.Toggle:
            component = pnw.Toggle(name="Toggle", button_type="success")
        elif self.component == pnw.VideoStream:
            component = pnw.VideoStream(
                name="Video Stream", sizing_mode="stretch_width", height=300
            )
        if not component:
            component = self.component(name="Hello")
        if not controls:
            controls = component.controls()
        controls.margin = 0
        self._component_panel[:] = [
            pn.pane.Markdown("## " + component.__class__.name + " " + self.component_type),
            component,
            pn.pane.Markdown("## Parameters"),
            controls,
        ]


def view():
    """Returns a small app for testing"""
    pn.config.sizing_mode = "stretch_width"
    return PanelComponentExplorer().view


if __name__.startswith("bokeh"):
    pn.config.sizing_mode = "stretch_width"
    app = PanelComponentExplorer()
    # pn.state.add_periodic_callback(app._update_css_panel, period=1000)
    app.view.servable()



"""The [NGL Viewer](https://github.com/nglviewer/ngl) can be used
to visualize and analyse pdb molecule structures.

Source: Discussion on [Discourse 583]\
(https://discourse.holoviz.org/t/how-to-use-ngl-webgl-protein-viewer-in-panel/583)

Author: https://github.com/Jhsmit
"""

import panel as pn
import param

NGL_JS = "https://unpkg.com/ngl@2.0.0-dev.37/dist/ngl.js"

REPRESENTATIONS = [
    "ball+stick",
    "backbone",
    "ball+stick",
    "cartoon",
    "hyperball",
    "licorice",
    "ribbon",
    "rope",
    "spacefill",
    "surface",
]
COLOR_SCHEMES = ["chainid", "residueindex", "chainname"]


class NGLViewer(pn.pane.HTML):
    """The [NGL Viewer](https://github.com/nglviewer/ngl) can be used
    to show and analyse pdb molecule structures"""

    pdb_string = param.String()
    rcsb_id = param.String()
    representation = param.Selector(
        default="ribbon",
        objects=REPRESENTATIONS,
    )
    color_scheme = param.Selector(default="chainid", objects=COLOR_SCHEMES)
    spin = param.Boolean(default=False)
    priority = 0
    _rename = dict(
        pn.pane.HTML._rename,
        pdb_string=None,
        rcsb_id=None,
        representation=None,
        spin=None,
        color_scheme=None,
    )

    def __init__(self, **params):
        if not "ngl_js" in pn.config.js_files:
            pn.config.js_files["ngl_js"] = NGL_JS

        super().__init__(**params)
        self.load_string = """
        stage = new NGL.Stage("viewport");
        stage.loadFile()"""
        self._update_object_from_parameters()

    @param.depends("representation", "spin", "color_scheme", watch=True)
    def _update_object_from_parameters(self):
        html = f"""\
<div id="viewport" style="width:100%; height:100%;"></div>
<script>
{self.load_string}.then(function(o){{
    o.addRepresentation("{self.representation}", {{colorScheme: "{self.color_scheme}"}});
    o.autoView();
    }}
);
stage.setSpin({'true' if self.spin else 'false'});
</script>
"""
        self.object = html

    @param.depends("pdb_string", watch=True)
    def _update_object_from_pdb_string(self):
        self.load_string = f"""
            var PDBString = `{self.pdb_string}`;
            stage = new NGL.Stage("viewport");
            stage.loadFile( new Blob([PDBString], {{type: 'text/plain'}}), {{ ext:'pdb'}} )"""
        self._update_object_from_parameters()

    @param.depends("rcsb_id", watch=True)
    def _update_object_from_rcsb_id(self):
        self.load_string = f"""
            stage = new NGL.Stage("viewport");
            stage.loadFile("rcsb://{self.rcsb_id}")"""
        self._update_object_from_parameters()



"""This module implements the ComponentReloader. The ComponentReloader is used by the Designer.
For each component you want access to in the Designer you should provide a seperate
Reload Service"""

import datetime
import importlib
import inspect
import pathlib
import sys
import traceback

import panel as pn
import param

from awesome_panel_extensions.developer_tools.designer.views import ErrorView


def _instance(parameter):
    if callable(parameter):
        return parameter()
    if inspect.isfunction(parameter) or inspect.isclass(parameter):
        return parameter()
    return parameter


def _to_instances(dictionary):
    return {parameter: _instance(value) for parameter, value in dictionary.items()}


def _to_parameterized(component_instance) -> param.Parameterized:
    """The component_instance should be an instance of param.Parameterized in order to be
    able to show it in the settings_pane via pn.Param. If not we wrap it with pn.panel"""
    if not isinstance(component_instance, param.Parameterized):
        component_instance = pn.panel(
            component_instance,
            min_height=400,
            sizing_mode="stretch_both",
        )
        if isinstance(component_instance, pn.pane.Plotly):
            component_instance.config = {"responsive": True}
        elif isinstance(component_instance, pn.pane.Vega):
            component_instance.margin = 25

    return component_instance


class ComponentReloader(param.Parameterized):  # pylint: disable=too-many-instance-attributes
    """The ComponentReloader is used by the Designer.
    For each component you want access to in the Designer you should provide a seperate
    Reload Service

    Args:
        component ([type]): For now the components that are know to be supported are

        - subclasses of `pn.reactive.Reactive`
        - subclasses of `param.Parameterized` with a `view` parameter which is a subclass of
        `pn.reactive.Reactive`

    Please NOTE that in order for the reload service to be able to reload the compoonent, the
    component specified cannot be defined in the __main__ file.

    Example
    -------

    ```python
    TITLE_COMPONENT = ComponentReloader(
        component=components.TitleComponent, css_path=COMPONENT_CSS, js_path=COMPONENT_JS,
    )
    EMPTY_COMPONENT = ComponentReloader(
        component=components.EmptyComponent, css_path=COMPONENT_CSS, js_path=COMPONENT2_JS,
    )
    ```"""

    component = param.Parameter(allow_None=False)
    parameters = param.Dict()
    component_instance = param.Parameter()
    css_path = param.Parameter(constant=True)
    js_path = param.Parameter(constant=True)
    modules_to_reload = param.List()

    reload_component = param.Action(label="RELOAD COMPONENT")
    reload_css_file = param.Action(label="RELOAD CSS")
    reload_js_file = param.Action(label="RELOAD JS")

    css_text = param.String()
    js_text = param.String()

    reloading = param.Boolean(default=False)
    last_reload = param.String(constant=True)
    error_message = param.String()

    def __init__(self, component, **params):
        if not isinstance(params, dict):
            params = {}
        params["component"] = component
        super().__init__(**params)

        try:
            name = self.component.name
        except AttributeError:
            name = self.component.__name__

        with param.edit_constant(self):
            self.name = name

        self.reload_component = self._reload_component
        self.reload_css_file = self._reload_css_file
        self.reload_js_file = self._reload_js_file

        self._parameter_instances = None

    def __repr__(self):
        return f"ComponentReloader({self.name})"

    def __str__(self):
        return f"ComponentReloader({self.name})"

    def _reload_component(self, _=None):
        try:
            self._signal_reload_start()

            if self.component_instance is not None:
                for mod in self.modules_to_reload:  # pylint: disable=not-an-iterable
                    importlib.reload(mod)

                mod = sys.modules[self.component.__module__]
                importlib.reload(mod)
                with param.edit_constant(self):
                    self.component = getattr(mod, self.component.__name__)
            if self.parameters:
                if not self._parameter_instances:
                    self._parameter_instances = _to_instances(self.parameters)

                # pylint: disable=not-a-mapping
                component_instance = self.component(**self._parameter_instances)
            else:
                component_instance = self.component()

            self.component_instance = _to_parameterized(component_instance)

            self._reset_error_message()
        except Exception as ex:  # pylint: disable=broad-except
            self._report_exception(ex)
        finally:
            self._signal_reload_end()

    def _reload_css_file(self, _=None):
        try:
            self._signal_reload_start()
            if not self.css_path:
                pass
            elif isinstance(self.css_path, pathlib.Path):
                self.css_text = self.css_path.read_text()
            else:
                raise NotImplementedError

            self._reset_error_message()
        except Exception as ex:  # pylint: disable=broad-except
            self._report_exception(ex)
        finally:
            self._signal_reload_end()

    def _reload_js_file(self, _=None):
        try:
            self._signal_reload_start()
            if not self.js_path:
                pass
            elif isinstance(self.js_path, pathlib.Path):
                self.js_text = self.js_path.read_text()
            else:
                raise NotImplementedError

            self._reset_error_message()
        except Exception as ex:  # pylint: disable=broad-except
            self._report_exception(ex)
        finally:
            self._signal_reload_end()

    def _signal_reload_start(self):
        self.reloading = True
        print("reload start", self.name, datetime.datetime.now())

    def _report_exception(self, ex):  # pylint: disable=unused-argument
        self.error_message = traceback.format_exc()
        self.component_instance = ErrorView(error_message=self.error_message)
        print(self.name, self.error_message)

    def _signal_reload_end(self):
        self.reloading = False
        with param.edit_constant(self):
            self.last_reload = str(datetime.datetime.now())
        print("reload end", self.name, datetime.datetime.now())

    def _reset_error_message(self):
        self.error_message = ""



"""The site module contains definitions of models like User and Application.

The models defined in this module can for example be used for

- Galleries
- Menus
- Governance. I.e. restricting access to certain applications for specific users.
- And much more.
"""
import uuid
from typing import List

import markdown
import panel as pn
import param
import yaml

from awesome_panel_extensions.assets.svg_icons import ICONS

AVATAR_URL = (
    "https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel-assets/master/awesome-panel/"
    "avatar.png"
)
MARKDOWN_EXTENSIONS = ["extra", "smarty", "codehilite"]
STYLE = """
img.pnx-avatar {
    height:100%;
    width:100%;
    border-radius:50%;
    vertical-align: middle;
}
.pnx-resource img.pnx-avatar {
    height: 2em;
    width: 2em;
    margin-left: 0.5em;
}
.pnx-resource svg.pnx-icon {
    height: 1.5em;
    margin-left: 0.5em;
    vertical-align: middle;
    fill: currentColor;
}
.pnx-resource svg.pnx-icon.pnx-icon-binder {
    width: 5em;
}
.pnx-resource a {
    text-decoration: none;
}
"""


def _get_nested_value(element, *keys):
    """Returns a nested value if it exists. Otherwise None

    If 'panel' is a key of element then we search from the value of 'panel'"""
    if not isinstance(element, dict):
        raise AttributeError("Expects dict as first argument.")
    if len(keys) == 0:
        raise AttributeError("Expects at least two arguments, one given.")

    if "panel" in element:
        element = element["panel"]

    if "config" in element:
        element = element["config"]

    for key in keys:
        try:
            element = element[key]
        except KeyError:
            return None
    return element


def _skip(item):
    """Returns True if the item is a dictionary with the key 'skip' and value True"""
    if not isinstance(item, dict):
        return False
    return item.get("skip", False)


class _BaseModel(param.Parameterized):
    """The BaseModel adds ordering by the name parameter to a Class"""

    uid = param.String(constant=True, doc="A unique id identifying the item.", precedence=1)
    name = param.String(default="New Model", doc="The name of the item", precedence=1)
    category = param.String(default="Other", doc="A category name", precedence=1)
    tags = param.List(
        class_=str,
        precedence=3,
        doc="""A list of tags like 'machine-learning', 'panel', 'holoviews'.""",
    )
    resources = param.Dict(
        precedence=3,
        doc="""
        A dictionary of urls. For example 'github': 'https://panel.holoviz.org'""",
    )

    def __init__(self, **params):
        if "uid" not in params:
            params["uid"] = str(uuid.uuid4())
        if "resources" not in params:
            params["resources"] = {}

        super().__init__(**params)

    def __lt__(self, other):
        if hasattr(other, "name"):
            return self.name.casefold() < other.name.casefold()
        return True

    def __eq__(self, other):
        if hasattr(other, "uid"):
            return self.uid == other.uid
        return False

    def __str__(
        self,
    ):
        return self.name

    def __repr__(
        self,
    ):
        return f"{self.__class__.name}(name='{self.name}')"


class User(_BaseModel):
    """A Model of a User

    >>> from awesome_panel_extensions.site.models import User
    >>> User(
    ...     name = "Philipp Rudiger",
    ...     url = "https://github.com/philippjfr",
    ...     email = "na@hotmail.com",
    ...     avatar = "https://avatars.githubusercontent.com/u/1550771",
    ...     tags = ["pyviz", "developer"],
    ...     resources = {"twitter": "https://twitter.com/PhilippJFR"},
    ... )
    User(name='Philipp Rudiger')
    """

    name = param.String(default="New User", doc="The name of the user.", precedence=1)

    email = param.String(doc="The email of the user.", precedence=1)
    url = param.String(doc="An url pointing to a page about the user.", precedence=1)

    avatar = param.String(
        default=AVATAR_URL, doc="The url of an avatar image of the user.", precedence=2
    )

    @staticmethod
    def _get_users(config):
        """Returns a list of Users from the specificed list of dicts"""
        if not config:
            return []

        users = _get_nested_value(config, "users")
        if not users:
            return []
        users = [{key: value for key, value in user.items() if value} for user in users]
        return [User(**user) for user in users if not _skip(user)]

    def _repr_html_(self):
        # pylint: disable=line-too-long
        return f"""<a href="{ self.url }" target="_blank"><img src="{ self.avatar }" class="pnx-avatar" alt="Avatar" title="{ self.name }"></a>"""


class Application(_BaseModel):
    """A Model of an Application

    >>> from awesome_panel_extensions.site.models import User, Application
    >>> philipp = User(
    ...     name = "Philipp Rudiger",
    ...     url = "https://github.com/philippjfr",
    ...     email = "na@hotmail.com",
    ...     avatar = "https://avatars.githubusercontent.com/u/1550771",
    ...     resources = {"twitter": "https://twitter.com/PhilippJFR"},
    ...     tags = ["pyviz", "developer"],
    ... )
    >>> Application(
    ...     name = "Panel",
    ...     description = "The analytics app framework to rule them all",
    ...     description_long = "Turns every dash into something lit",
    ...     author = philipp,
    ...     owner = philipp,
    ...     url = "https://panel.holoviz.org",
    ...     thumbnail = "https://panel.holoviz.org/_static/logo_stacked.png",
    ...     resources = {"github": "https://github.com/holoviz/panel"},
    ...     tags = ["awesome", "analytics", "apps"]
    ... )
    Application(name='Panel')
    """

    name = param.String(
        default="New Application",
        precedence=1,
        doc="""
        The name of the application""",
    )

    author = param.ClassSelector(class_=User, constant=True, precedence=1)
    owner = param.ClassSelector(class_=User, constant=True, precedence=1)

    description = param.String(
        regex="^.{0,150}$",
        precedence=1,
        doc="""
        A short text introduction of max 150 characters.""",
    )
    description_long = param.String(
        precedence=1,
        doc="""
        A longer description. Can contain Markdown and HTML""",
    )
    project = param.String(
        doc="""
    The name of associated project. Can be used for governance in a multi-project site""",
        precedence=1,
    )
    servable = param.String(precedence=2, doc="The path to a servable Panel application")
    url = param.String(precedence=2, doc="The url of the application.")
    thumbnail = param.String(precedence=2, doc="The url of a thumbnail of the application.")

    def __init__(self, **params):
        for key in ["author", "owner"]:
            if key in params:
                user = params[key]
                if isinstance(user, str):
                    params[key] = User(uid=user, name=user)
            else:
                params[key] = User()

        super().__init__(**params)

        if not STYLE in pn.config.raw_css:
            pn.config.raw_css.append(STYLE)

    @staticmethod
    def _get_applications(config, users):
        """Returns a list of Applications from the specificed list of dicts.

        users is a list of existing Users
        """
        if not config:
            return []

        user_map = {user.uid: user for user in users}

        applications = _get_nested_value(config, "applications")
        if not applications:
            return []

        for application in applications:
            for key in ["author", "owner"]:
                if key in application:
                    user = application[key]
                    if user in user_map:
                        application[key] = user_map[user]
        applications = [
            {key: value for key, value in application.items() if value}
            for application in applications
        ]
        return [
            Application(**application) for application in applications if not _skip(application)
        ]

    @classmethod
    def read(cls, file) -> List["Application"]:
        """Returns a list of Applications from the specified file

        Currently only yaml is supported

        Args:
            path (str|pathlib.Path): The path to the file
        Returns:
            List[Application]: A list of applications
        """
        with open(file, "r") as stream:
            config = yaml.safe_load(stream)
        users = User._get_users(config)  # pylint: disable=protected-access
        return cls._get_applications(config, users)

    @staticmethod
    def _get_url_icon_html(
        title,
        url,
    ):
        return (
            f"""<a title="{ title }" appearance="hypertext" href="{ url }" target="_blank">"""
            f"""{ ICONS[title] }</a>"""
        )

    @staticmethod
    def _markdown_to_html(text: str) -> str:
        return markdown.markdown(text, extensions=MARKDOWN_EXTENSIONS, output_format="html5")

    def intro_section(self) -> pn.pane.HTML:
        """An panel with a text introduction to the Resource

        Returns:
            pn.pane.HTML: The Intro Section panel.
        """
        return pn.pane.HTML(self._repr_html_())

    def _repr_html_(self):
        description = self._markdown_to_html(self.description_long)
        html = f"""<div class="pnx-resource">
        <h1 class="pnx-header">{ self.name }</h1>
        <p>{ description }</p>
        """
        if self.author:
            # pylint: disable=protected-access
            html += f"""<p><strong>Authors:</strong>{ self.author._repr_html_() }</p>"""

        code_url = self.resources.get("code", "")
        if code_url:
            code = self._get_url_icon_html("code", code_url)
            html += "<p><strong>Code:</strong>" + code + "</p>"
        resources = ""
        for item in ["doc", "gif", "mp4", "youtube", "binder"]:
            url = self.resources.get(item, "")
            if url:
                resources += self._get_url_icon_html(item, url)

        if resources:
            html += "<p><strong>Resources:</strong>" + resources + "</p>"

        tags = ", #".join(self.tags)
        if tags:
            html += "<p><strong>Tags:</strong> #" + tags.lower() + "</p>"

        html += "</div>"
        return html



"""Implementation of MWC Material Button"""
import param

from awesome_panel_extensions.frameworks._base.config import BUTTON_TYPES
from awesome_panel_extensions.frameworks.material.config import MWC_ICONS
from awesome_panel_extensions.web_component import WebComponent

# pylint: disable=abstract-method


class Button(WebComponent):
    """Implementation of mwc-button

    Set the `name` to set the text shown to the user
    Set the `icon` to set the icon shown to the user

    Set `unelevated` or `raised` to change the style
    """

    html = param.String("<mwc-button style='width:100%'></mwc-button")
    attributes_to_watch = param.Dict(
        {
            "label": "name",
            "icon": "icon",
            "raised": "raised",
            "unelevated": "unelevated",
            "disabled": "disabled",
            "trailingIcon": "trailing_icon",
            "class": "_button_css_class",
        }
    )
    events_to_watch = param.Dict({"click": "clicks"})

    button_type = param.ObjectSelector(default="default", objects=BUTTON_TYPES)
    disabled = param.Boolean(default=False)

    icon = param.ObjectSelector(default=None, objects=MWC_ICONS, allow_None=True)
    trailing_icon = param.Boolean(default=False)
    raised = param.Boolean(default=False)
    unelevated = param.Boolean(default=False)

    clicks = param.Integer()

    _button_css_class = param.String()

    height = param.Integer(default=30)

    def __init__(self, **params):
        super().__init__(**params)

        if self.button_type != "default":
            self._handle_button_type_change()

    @param.depends("button_type", watch=True)
    def _handle_button_type_change(self, *_):
        if self.button_type in ["primary", "success", "warning", "danger"]:
            self.param.set_param(raised=True, unelevated=False, _button_css_class=self.button_type)
        else:
            self.param.set_param(
                raised=self.param.raised.default,
                unelevated=self.param.unelevated.default,
                _button_css_class=self.button_type,
            )



"""The FastTextInput extends the Panel TextInput to a Fast Design Framework TextInput.

It is built on the the fast-text-field web component. The component supports two visual apperances
(outline and filled).

For more information view the [component specification]\
(https://github.com/microsoft/fast/tree/master/packages/web-components/fast-foundation/src/\
text-field).

See also https://explore.fast.design/components/fast-text-field.
    """
# Note:
# rename to fast_text_input in accordance with bokeh text_input
# rename maxlength, minlength in accordance with bokeh text_input
# in the .ts set maxLength, minLength instead of maxlength, minlength.

import panel as pn
import param  # pylint: disable=wrong-import-order
from panel.widgets import TextInput

from awesome_panel_extensions.bokeh_extensions.fast.fast_text_input import (
    FastTextInput as _BkFastTextInput,
)

FAST_TEXT_INPUT_APPEARENCES = [
    "outline",
    "filled",
]
DEFAULT_TEXT_INPUT_APPEARANCE = "outline"
TYPES = ["email", "password", "tel", "text", "url"]
DEFAULT_TYPE = "text"


class _FastTextInputMixin(pn.widgets.Widget):
    # To used by FastTextInput and FastLiteralInput

    # value  is inherited
    # placeholder is inherited
    # list is not supported
    placeholder = param.String(
        default="",
        doc="A placeholder string displayed when no value is entered.",
    )
    appearance = param.ObjectSelector(
        default=DEFAULT_TEXT_INPUT_APPEARANCE,
        objects=FAST_TEXT_INPUT_APPEARENCES,
        doc="""Determines the appearance of the textinput. One of `outline` or `filled`.
        Defaults to outlined""",
        allow_None=True,
    )
    autofocus = param.Boolean(
        default=False,
        doc="""The autofocus attribute. Defaults to `False`""",
    )
    type_of_text = param.ObjectSelector(
        default=DEFAULT_TYPE,
        objects=TYPES,
        label="Type",
        doc="""Determines the type of text accepted. One of `email`, `password`, `tel`, `text` or
        `url`. Defaults to text.
        """,
    )
    max_length = param.Integer(
        default=100,
        doc="""The maximum length of the text string""",
    )
    min_length = param.Integer(
        default=0,
        doc="""The minimum length of the text string""",
    )
    pattern = param.String(
        default=None,
        doc="""A regular expression that the input's value must match in order for the value to
        pass constraint validation""",
    )
    # Cannot get size working
    # It raises an error and in the Fast Component Explorer I cannot see any effect of this
    # attribute
    # size = param.Integer(
    #     default=None,
    #     doc="""Valid for email, password, tel, and text input types only. Specifies how much of
    # the input is shown""",
    #     allow_None=True,
    # )
    spellcheck = param.Boolean(
        default=False, doc="""Whether or not the spell check is enabled. Default is False"""
    )
    required = param.Boolean(
        default=False, doc="""Whether or not the FastTextInput is required. Default is False"""
    )
    disabled = param.Boolean(
        default=False, doc="""Whether or not the FastTextInput is disabled. Default is False"""
    )
    readonly = param.Boolean(
        default=False, doc="""Whether or not the FastTextInput is readonly. Default is False"""
    )

    # If we don't set this then the pane below will overlap
    height = param.Integer(default=60, bounds=(0, None))

    _widget_type = _BkFastTextInput

    _rename = {
        **pn.widgets.TextInput._rename,  # pylint: disable=protected-access
    }


class FastTextInput(_FastTextInputMixin, TextInput):
    """The FastTextInput extends the Panel TextInput into the Fast Design Framework.

It is built on the the fast-textinput web component.

For more information view the [component specification]\
(https://github.com/microsoft/fast/tree/master/packages/web-components/fast-foundation/src/\
text-field).

See also https://explore.fast.design/components/fast-text-field.
    """



"""The NBViewerLinkButton displayes the Binder badge and if clicked opens the Notebook on Binder
    in a new tab"""
import panel as pn
import param

from awesome_panel_extensions.widgets.link_buttons.image_link_button import DerivedImageLinkButton

IMAGE_URL = (
    "https://raw.githubusercontent.com/jupyter/design/master/logos/Badges/nbviewer_badge.svg"
)


class NBViewerLinkButton(DerivedImageLinkButton):
    """The NBViewerLinkButton displayes the Binder badge and if clicked opens the Notebook on Binder
    in a new tab"""

    repository = param.String()
    branch = param.String()
    folder = param.String()
    notebook = param.String()

    image_url = param.String(default=IMAGE_URL, doc="The url to the image", constant=True)

    width = param.Integer(
        default=170,
        bounds=(0, None),
        doc="""
        The width of the component (in pixels). This can be either
        fixed or preferred width, depending on width sizing policy.""",
    )

    def __init__(self, **params):
        super().__init__(**params)
        self.style = {}
        self._update_link_url_from_parameters()

    @param.depends(
        "repository", "branch", "folder", "notebook", "height", "width", "sizing_mode", watch=True
    )
    def _update_link_url_from_parameters(self, *events):  # pylint: disable=unused-argument
        with param.edit_constant(self):
            self.link_url = (
                f"https://nbviewer.jupyter.org/github/{self.repository}/blob/"
                f"{self.branch}/{self.folder}/{self.notebook}"
            )


if __name__.startswith("bokeh"):
    button = NBViewerLinkButton(
        repository="MarcSkovMadsen/awesome-panel-extensions",
        branch="master",
        folder="examples/panes",
        notebook="WebComponent.ipynb",
    )
    settings_pane = pn.Param(
        button,
        parameters=[
            "repository",
            "branch",
            "folder",
            "notebook",
            "height",
            "width",
            "sizing_mode",
            "margin",
        ],
        background="lightgray",
        sizing_mode="stretch_width",
    )
    app = pn.Column(button, settings_pane, width=500, height=800)
    app.servable()
